// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OCI.BiDir;

//
// ActiveTransport encapsulates a "real" Transport and is shared
// between two virtual (i.e., forward and callback) connections
//
class ActiveTransport
{
    //
    // The magic bytes we expect at the beginning of each message
    //
    private static final byte[] magic_ =
        { (byte)'O', (byte)'O', (byte)'C', (byte)'B' };

    //
    // The local peer id
    //
    private String localPeer_;

    //
    // The remote peer id
    //
    private String remotePeer_;

    //
    // The transport observer
    //
    private TransportObserver observer_;

    //
    // The transport delegate
    //
    private com.ooc.OCI.Transport delegate_;

    //
    // Have we seen any messages on the "callback" connection?
    //
    private boolean callbackMessage_;

    //
    // Incoming buffers on the "forward" connection
    //
    private java.util.Vector incoming_ = new java.util.Vector();
    private java.lang.Object incomingCond_ = new java.lang.Object();

    //
    // Incoming buffers on the "callback" connection
    //
    private java.util.Vector incomingCallback_ = new java.util.Vector();
    private java.lang.Object incomingCallbackCond_ = new java.lang.Object();

    //
    // The SystemException detected by the receiver thread
    //
    private org.omg.CORBA.SystemException receiveEx_;

    //
    // The receiver thread
    //
    private static class ReceiverThread extends Thread
    {
        private ActiveTransport activeTransport_;

        ReceiverThread(ActiveTransport activeTransport)
        {
            super("ORBacus:BiDirIOP:Transport:ReceiverThread");
            activeTransport_ = activeTransport;
        }

        public void
        run()
        {
            activeTransport_.receiverRun();

            //
            // Break cyclic object dependency
            //
            activeTransport_ = null;
        }
    }
    private ReceiverThread receiverThread_;

    //
    // The sender thread
    //
    private static class SenderThread extends Thread
    {
        private ActiveTransport activeTransport_;

        SenderThread(ActiveTransport activeTransport)
        {
            super("ORBacus:BiDirIOP:Transport:SenderThread");
            activeTransport_ = activeTransport;
        }

        public void
        run()
        {
            activeTransport_.senderRun();

            //
            // Break cyclic object dependency
            //
            activeTransport_ = null;
        }
    }
    private SenderThread senderThread_;

    //
    // Outgoing messages
    //
    private java.util.Vector outgoing_ = new java.util.Vector();
    private java.lang.Object outgoingMonitor_ = new java.lang.Object();

    //
    // Have we been shutdown?
    //
    private boolean shutdown_;

    //
    // Was this transport created for an outgoing or incoming connection?
    //
    static final int Outgoing = 0;
    static final int Incoming = 1;
    private int type_;

    // ------------------------------------------------------------------
    // Private member implementations
    // ------------------------------------------------------------------

    //
    // Create a buffer containing the BiDir message header
    //
    private com.ooc.OCI.Buffer
    createHeader(boolean callback, int size)
    {
        com.ooc.BiDirIOP.Header h = new com.ooc.BiDirIOP.Header();
        h.magic = magic_;
        h.bidiriop_version = new com.ooc.BiDirIOP.Version((byte)1, (byte)0);
        h.flags = 0; // big endian (JVM is always big endian)
        if(callback)
            h.flags |= 0x2;
        h.reserved = new byte[5];
        h.message_size = size;

        com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer();
        com.ooc.CORBA.OutputStream out = new com.ooc.CORBA.OutputStream(buf);
        com.ooc.BiDirIOP.HeaderHelper.write(out, h);
        buf.pos(0);

        return buf;
    }

    //
    // Extract a BiDir message header from the buffer
    //
    private com.ooc.BiDirIOP.Header
    extractHeader(com.ooc.OCI.Buffer buf)
    {
        com.ooc.CORBA.InputStream in =
            new com.ooc.CORBA.InputStream(buf);
        // TODO: Unmarshal manually?
        com.ooc.BiDirIOP.Header h =
            com.ooc.BiDirIOP.HeaderHelper.read(in);

        //
        // Check byte order and re-read header if we need to swap
        //
        if((h.flags & 0x1) != 0)
        {
            in = new com.ooc.CORBA.InputStream(buf, 0, true);
            h = com.ooc.BiDirIOP.HeaderHelper.read(in);
        }

        if(h.magic[0] != magic_[0] || h.magic[1] != magic_[1] ||
           h.magic[2] != magic_[2] || h.magic[3] != magic_[3])
            throw new org.omg.CORBA.COMM_FAILURE("Bad magic key");

        return h;
    }

    //
    // Fill as much of the destination buffer as possible
    // (not mutex protected)
    //
    private void
    fillBuffer(com.ooc.OCI.Buffer dst, boolean callback)
    {
        java.util.Vector vec;
        if(callback)
            vec = incomingCallback_;
        else
            vec = incoming_;

        //
        // Continue until we've filled the buffer or until there are no
        // more incoming buffers available
        //
        while(!dst.is_full() && !vec.isEmpty())
        {
            com.ooc.OCI.Buffer src = (com.ooc.OCI.Buffer)vec.elementAt(0);
            int dstLen = dst.rest_length();
            int srcLen = src.rest_length();
            int len = (srcLen > dstLen ? dstLen : srcLen);
            System.arraycopy(src.data(), src.pos(),
                             dst.data(), dst.pos(), len);
            dst.advance(len);
            src.advance(len);
            if(src.rest_length() == 0)
                vec.removeElementAt(0);
        }
    }

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    //
    // Constructor for an outgoing connection. Both the local and remote
    // peers are known.
    //
    ActiveTransport(String localPeer, String remotePeer,
                    TransportObserver observer,
                    com.ooc.OCI.Transport delegate)
    {
        localPeer_ = localPeer;
        remotePeer_ = remotePeer;
        observer_ = observer;
        delegate_ = delegate;
        callbackMessage_ = false;
        receiveEx_ = null;
        shutdown_ = false;
        type_ = Outgoing;

        //
        // Start receiver thread
        //
        receiverThread_ = new ReceiverThread(this);
        receiverThread_.start();

        //
        // Start sender thread
        //
        senderThread_ = new SenderThread(this);
        senderThread_.start();
    }

    //
    // Constructor for an incoming connection. Only the local peer
    // is known at construction.
    //
    ActiveTransport(String localPeer, TransportObserver observer,
                    com.ooc.OCI.Transport delegate)
    {
        localPeer_ = localPeer;
        remotePeer_ = null;
        observer_ = observer;
        delegate_ = delegate;
        callbackMessage_ = false;
        receiveEx_ = null;
        shutdown_ = false;
        type_ = Incoming;

        //
        // Start receiver thread
        //
        receiverThread_ = new ReceiverThread(this);
        receiverThread_.start();

        //
        // Start sender thread
        //
        senderThread_ = new SenderThread(this);
        senderThread_.start();
    }

    // ------------------------------------------------------------------
    // Package member implementations
    // ------------------------------------------------------------------

    String
    localPeer()
    {
        return localPeer_;
    }

    String
    remotePeer()
    {
        return remotePeer_;
    }

    int
    type()
    {
        return type_;
    }

    com.ooc.OCI.Transport
    delegate()
    {
        return delegate_;
    }

    com.ooc.OCI.TransportInfo
    info()
    {
        return delegate_.get_info();
    }

    void
    close(boolean callback)
    {
        if(callback)
        {
            synchronized(incomingCallbackCond_)
            {
                //
                // Notify any threads waiting to receive from the
                // callback direction, so that connection loss can
                // be detected
                //
                incomingCallbackCond_.notifyAll();

                //
                // If no more messages have been received on the
                // callback connection, then reset callbackMessage_.
                // Otherwise, inform the observer that we have
                // more callback messages waiting.
                //
                if(incomingCallback_.isEmpty())
                    callbackMessage_ = false;
                else
                    observer_.callbackConnection(this);
            }
        }
        else
        {
            observer_.shutdown(this);

            synchronized(outgoingMonitor_)
            {
                if(!shutdown_)
                {
                    shutdown_ = true;
                    outgoingMonitor_.notify();
                }
            }

            delegate_.close();
        }
    }

    void
    shutdown(boolean callback)
    {
        if(callback)
        {
            //
            // Notify any threads waiting to receive from the
            // callback direction, so that connection loss can
            // be detected
            //
            synchronized(incomingCallbackCond_)
            {
                incomingCallbackCond_.notifyAll();
            }
        }
        else
        {
            observer_.shutdown(this);

            synchronized(outgoingMonitor_)
            {
                shutdown_ = true;
                outgoingMonitor_.notify();
            }

            delegate_.shutdown();
        }
    }

    void
    receive(com.ooc.OCI.Buffer buf, boolean block, boolean callback,
            ShutdownStatus status)
    {
        java.lang.Object mutex;
        if(callback)
            mutex = incomingCallbackCond_;
        else
            mutex = incomingCond_;

        synchronized(mutex)
        {
            fillBuffer(buf, callback);

            if(!buf.is_full() && receiveEx_ != null)
                throw receiveEx_;

            if(block)
            {
                while(!buf.is_full())
                {
                    try
                    {
                        mutex.wait();
                    }
                    catch(InterruptedException ex)
                    {
                    }

                    fillBuffer(buf, callback);

                    if(!buf.is_full())
                    {
                        if(receiveEx_ != null)
                            throw receiveEx_;

                        if(status != null && status.isShutdown())
                        {
                            throw new org.omg.CORBA.COMM_FAILURE(
                                com.ooc.OB.MinorCodes.describeCommFailure(
                                    com.ooc.OB.MinorCodes.MinorRecv),
                                com.ooc.OB.MinorCodes.MinorRecv,
                                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
                        }
                    }
                }
            }
        }
    }

    boolean
    receiveDetect(com.ooc.OCI.Buffer buf, boolean block, boolean callback,
                  ShutdownStatus status)
    {
        java.lang.Object mutex;
        if(callback)
            mutex = incomingCallbackCond_;
        else
            mutex = incomingCond_;

        synchronized(mutex)
        {
            fillBuffer(buf, callback);

            if(!buf.is_full() && receiveEx_ != null)
                return false;

            if(block)
            {
                while(!buf.is_full())
                {
                    try
                    {
                        mutex.wait();
                    }
                    catch(InterruptedException ex)
                    {
                    }

                    fillBuffer(buf, callback);

                    if(!buf.is_full())
                    {
                        if(receiveEx_ != null)
                            return false;

                        if(status != null && status.isShutdown())
                            return false;
                    }
                }
            }
        }

        return true;
    }

    void
    receiveTimeout(com.ooc.OCI.Buffer buf, int t, boolean callback,
                   ShutdownStatus status)
    {
        java.lang.Object mutex;
        if(callback)
            mutex = incomingCallbackCond_;
        else
            mutex = incomingCond_;

        synchronized(mutex)
        {
            fillBuffer(buf, callback);

            if(!buf.is_full() && receiveEx_ != null)
                throw receiveEx_;

            long timeDone = System.currentTimeMillis() + t;
            while(!buf.is_full())
            {
                //
                // Have we timed out?
                //
                long now = System.currentTimeMillis();
                if(timeDone <= now)
                    break;

                //
                // No, calculate how long we have to wait now
                //
                long interval = timeDone - now;

                //
                // Block until we are notified, or until we timeout
                //
                try
                {
                    mutex.wait(interval);
                }
                catch(InterruptedException ex)
                {
                }

                fillBuffer(buf, callback);

                if(!buf.is_full())
                {
                    if(receiveEx_ != null)
                        throw receiveEx_;

                    if(status != null && status.isShutdown())
                    {
                        throw new org.omg.CORBA.COMM_FAILURE(
                            com.ooc.OB.MinorCodes.describeCommFailure(
                                com.ooc.OB.MinorCodes.MinorRecv),
                            com.ooc.OB.MinorCodes.MinorRecv,
                            org.omg.CORBA.CompletionStatus.COMPLETED_NO);
                    }
                }
            }
        }
    }

    boolean
    receiveTimeoutDetect(com.ooc.OCI.Buffer buf, int t, boolean callback,
                         ShutdownStatus status)
    {
        java.lang.Object mutex;
        if(callback)
            mutex = incomingCallbackCond_;
        else
            mutex = incomingCond_;

        synchronized(mutex)
        {
            fillBuffer(buf, callback);

            if(!buf.is_full() && receiveEx_ != null)
                return false;

            long timeDone = System.currentTimeMillis() + t;
            while(!buf.is_full())
            {
                //
                // Have we timed out?
                //
                long now = System.currentTimeMillis();
                if(timeDone <= now)
                    break;

                //
                // No, calculate how long we have to wait now
                //
                long interval = timeDone - now;

                //
                // Block until we are notified, or until we timeout
                //
                try
                {
                    mutex.wait(interval);
                }
                catch(InterruptedException ex)
                {
                }

                fillBuffer(buf, callback);

                if(!buf.is_full())
                {
                    if(receiveEx_ != null)
                        return false;

                    if(status != null && status.isShutdown())
                        return false;
                }
            }
        }

        return true;
    }

    void
    send(com.ooc.OCI.Buffer buf, boolean block, boolean callback)
    {
        if(shutdown_)
            throw new org.omg.CORBA.COMM_FAILURE();

        //
        // Create BiDir message and add to outgoing list
        //
        com.ooc.OCI.Buffer hbuf = createHeader(callback, buf.rest_length());
        Message msg = new Message(hbuf, buf);
        synchronized(outgoingMonitor_)
        {
            if(shutdown_)
                throw new org.omg.CORBA.COMM_FAILURE();

            outgoing_.addElement(msg);
            outgoingMonitor_.notify();
        }

        msg.waitUntilSent(block);
    }

    boolean
    sendDetect(com.ooc.OCI.Buffer buf, boolean block, boolean callback)
    {
        if(shutdown_)
            return false;

        //
        // Create BiDir message and add to outgoing list
        //
        com.ooc.OCI.Buffer hbuf = createHeader(callback, buf.rest_length());
        Message msg = new Message(hbuf, buf);
        synchronized(outgoingMonitor_)
        {
            if(shutdown_)
                return false;

            outgoing_.addElement(msg);
            outgoingMonitor_.notify();
        }

        try
        {
            msg.waitUntilSent(block);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            return false;
        }

        return true;
    }

    void
    sendTimeout(com.ooc.OCI.Buffer buf, int t, boolean callback)
    {
        if(shutdown_)
            throw new org.omg.CORBA.COMM_FAILURE();

        //
        // Create BiDir message and add to outgoing list
        //
        com.ooc.OCI.Buffer hbuf = createHeader(callback, buf.rest_length());
        Message msg = new Message(hbuf, buf);
        synchronized(outgoingMonitor_)
        {
            if(shutdown_)
                throw new org.omg.CORBA.COMM_FAILURE();

            outgoing_.addElement(msg);
            outgoingMonitor_.notify();
        }

        msg.waitUntilSent(t);
    }

    boolean
    sendTimeoutDetect(com.ooc.OCI.Buffer buf, int t, boolean callback)
    {
        if(shutdown_)
            return false;

        //
        // Create BiDir message and add to outgoing list
        //
        com.ooc.OCI.Buffer hbuf = createHeader(callback, buf.rest_length());
        Message msg = new Message(hbuf, buf);
        synchronized(outgoingMonitor_)
        {
            if(shutdown_)
                return false;

            outgoing_.addElement(msg);
            outgoingMonitor_.notify();
        }

        try
        {
            msg.waitUntilSent(t);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            return false;
        }

        return true;
    }

    void
    sendLocalPeerId()
    {
        com.ooc.OB.Assert._OB_assert(type_ == Outgoing);

        //
        // Transmit the peer id. The message format is
        //
        //   boolean endian
        //   sequence<octet> id
        //
        com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer();
        com.ooc.CORBA.OutputStream out = new com.ooc.CORBA.OutputStream(buf);
        out._OB_writeEndian();
        int len = localPeer_.length();
        out.write_ulong(len);
        byte[] bytes = new byte[len];
        for(int i = 0 ; i < len ; i++)
            bytes[i] = (byte)localPeer_.charAt(i);
        out.write_octet_array(bytes, 0, len);
        buf.pos(0);
        send(buf, true, false);
    }

    void
    receiveRemotePeerId()
    {
        com.ooc.OB.Assert._OB_assert(type_ == Incoming);

        //
        // Receive the peer id. The message format is
        //
        //   boolean endian
        //   sequence<octet> id
        //
        com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer(8);
        receive(buf, true, false, null);
        com.ooc.OB.Assert._OB_assert(buf.is_full());
        com.ooc.CORBA.InputStream in = new com.ooc.CORBA.InputStream(buf);
        in._OB_readEndian();
        int len = in.read_ulong();
        buf.alloc(len);
        receive(buf, true, false, null);
        com.ooc.OB.Assert._OB_assert(buf.is_full());
        remotePeer_ = new String(buf.data(), 0, buf.length());
    }

    //
    // Main loop for the receiver thread
    //
    public void
    receiverRun()
    {
        try
        {
            com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer();
            while(true)
            {
                //
                // Read header
                //
                buf.alloc(16);
                delegate_.receive(buf, true);
                com.ooc.OB.Assert._OB_assert(buf.is_full());
                com.ooc.BiDirIOP.Header hdr = extractHeader(buf);

                //
                // Read body
                //
                buf.alloc(hdr.message_size);
                delegate_.receive(buf, true);
                com.ooc.OB.Assert._OB_assert(buf.is_full());

                //
                // Consume buffer
                //
                com.ooc.OCI.Buffer b = new com.ooc.OCI.Buffer();
                b.consume(buf);
                b.pos(0);

                //
                // Add the buffer to the appropriate list and notify any
                // waiters
                //
                if((hdr.flags & 0x2) == 0x2)
                {
                    //
                    // Callback message
                    //
                    synchronized(incomingCallbackCond_)
                    {
                        //
                        // Is this the first callback message?
                        //
                        if(!callbackMessage_)
                        {
                            callbackMessage_ = true;
                            observer_.callbackConnection(this);
                        }

                        incomingCallback_.addElement(b);
                        incomingCallbackCond_.notifyAll();
                    }
                }
                else
                {
                    synchronized(incomingCond_)
                    {
                        incoming_.addElement(b);
                        incomingCond_.notifyAll();
                    }
                }
            }
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            receiveEx_ = ex;

            //
            // Notify any waiters that an exception occurred
            //
            synchronized(incomingCond_)
            {
                incomingCond_.notifyAll();
            }
            synchronized(incomingCallbackCond_)
            {
                incomingCallbackCond_.notifyAll();
            }
        }
    }

    //
    // Main loop for the sender thread
    //
    public void
    senderRun()
    {
        try
        {
            while(!shutdown_)
            {
                Message msg = null;

                synchronized(outgoingMonitor_)
                {
                    while(!shutdown_ && outgoing_.isEmpty())
                    {
                        try
                        {
                            outgoingMonitor_.wait();
                        }
                        catch(InterruptedException ex)
                        {
                        }
                    }

                    if(!shutdown_)
                    {
                        msg = (Message)outgoing_.firstElement();
                        outgoing_.removeElementAt(0);
                    }
                }

                if(msg != null)
                {
                    try
                    {
                        com.ooc.OCI.Buffer header = msg.header();
                        delegate_.send(header, true);
                        com.ooc.OB.Assert._OB_assert(header.is_full());

                        com.ooc.OCI.Buffer body = msg.body();
                        delegate_.send(body, true);
                        com.ooc.OB.Assert._OB_assert(body.is_full());

                        msg.sent();
                    }
                    catch(org.omg.CORBA.SystemException ex)
                    {
                        msg.exception(ex);
                        throw ex;
                    }
                }
            }

            //
            // We're shutdown
            //
            synchronized(outgoingMonitor_)
            {
                for(int i = 0 ; i < outgoing_.size() ; i++)
                {
                    Message m = (Message)outgoing_.elementAt(i);
                    m.exception(new org.omg.CORBA.COMM_FAILURE());
                }
            }
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            synchronized(outgoingMonitor_)
            {
                for(int i = 0 ; i < outgoing_.size() ; i++)
                {
                    Message m = (Message)outgoing_.elementAt(i);
                    m.exception(ex);
                }
            }
        }
    }
}
