#!/usr/bin/perl

#
# If strict stubs are used then an error occurs under UNIX:
# Bareword "INFINITE" not allowed while "strict subs" in use at
# ./PerformanceTest.pl line 206.
#
use strict;
use lib '../test';
use TestSuiteUtilities;

#
# The default set of data files
#
opendir(DATADIR, 'data');
my @DataFiles = sort (grep { /.*\.in/ } readdir(DATADIR));
closedir(DATADIR);

for(@DataFiles)
{
    $_ = "data/$_";
}

#
# Filename extension for test results.
#
my $OUT = '.out';

my $doRunTest = 1;
my $useExistingService = 0;
my $channelRef = "";
my $verbose = 0;

while($#ARGV >= 0)
{
    if($ARGV[0] =~ /[-]+[h?]/)
    {
	print "$0 [-h?] [-n] [-v] [-f reffile] [data file 1] [data file 2]" .
		"...\n";
	print "    -h, -? displays command line options\n";
	print "    -n     processes test data without running any tests\n";
	print "    -v     enables verbose mode\n";
	print "    -f     use channel IOR from file\n";
	exit 0;
    }
    elsif($ARGV[0] =~ /[-]+[f?]/)
    {
	$useExistingService = 1;
	shift @ARGV;

	my $channelRefFile = $ARGV[0];

	open(REFFILE, $channelRefFile) ||
	    die "Could not open $channelRefFile\n";
	$channelRef = <REFFILE>;
	close(REFFILE);

	print "channel ref: $channelRef\n";

	shift @ARGV;
    }
    elsif($ARGV[0] =~ /[-]+[v?]/)
    {
	$verbose = 1;
	shift @ARGV;
    }
    elsif($ARGV[0] =~ /[-]+[n?]/)
    {
	#
	# Don't actually run any tests, simply produce nice output
	#
	$doRunTest = 0;
	shift @ARGV;
    }
    else
    {
	if($#ARGV >= 0)
	{
	    #
	    # Use the files supplied on the command line
	    #
	    @DataFiles = @ARGV;
	}
        last;
    }
}

&initializeTools();

#
# Append the perf classes to the CLASSPATH
#            
if($^O eq "MSWin32")
{
    $ENV{"CLASSPATH"}="../perf/classes;" . $ENV{"CLASSPATH"};
}
else
{
    $ENV{"CLASSPATH"}="../perf/classes:" . $ENV{"CLASSPATH"};
}

if($useExistingService == 1)
{
    &useExistingService();
}

#
# Open a log file
#
open(LOG, ">perf.log");

#
# Run the tests.
#
foreach my $dataFile (@DataFiles)
{
    &printBanner($dataFile);

    if($doRunTest)
    {
	&runTest($dataFile);
    }

    &postProcess($dataFile);
}

close(LOG);

#
# Run a specific test
#
sub runTest($)
{
    my $dataFile = shift;
    my $java = "java -Xbootclasspath/p:../../ob/lib";
    my $config = "-ORBconfig ob.conf";

    my $EXE = "";
    if($^O eq "MSWin32")
    {
        $EXE = ".exe";
    }

    #
    # Hash of command line options for the supplier, consumer, and admin
    #
    my %perfOpts =
    (
	'numConsumers', '',	# integer
	'numSuppliers', '',	# integer
	'eventsPerSupplier', ''	# integer
    );

    open(DATAFIL, $dataFile) || die "Could not open $dataFile\n";
    while(<DATAFIL>)
    {
	chomp;
	my @param = split(":");
	$perfOpts{$param[0]} = $param[1];
    }
    close DATAFIL;

    #
    # Start a fresh Event Service
    #
    # TODO: implement a smarter startService that ensures the service
    # is ready.
    #
    my $eventPID = &startService();
    sleep(5);

    #
    # Time the number of events specified
    #
    my $totalSupplierEvents = $perfOpts{'eventsPerSupplier'};
    my $totalConsumerEvents = $totalSupplierEvents * $perfOpts{'numSuppliers'};
    my $totalEvents = $totalConsumerEvents * $perfOpts{'numConsumers'};

    #
    # Consumer command line
    #
    my $consumerExe = "$java event.Consumer $config";
    my $consumerArgs = " --total-events $totalConsumerEvents";

    #
    # Supplier command line
    #
    my $supplierExe = "$java event.Supplier $config";
    my $supplierArgs = " --total-events $totalSupplierEvents";

    if($verbose == 1)
    {
	$consumerArgs .= " --verbose";
	$supplierArgs .= " --verbose";
    }

    if($channelRef)
    {
	$consumerArgs .= " --channel-ref $channelRef";
	$supplierArgs .= " --channel-ref $channelRef";
    }

    my $outDataFile = getOutputFileName($dataFile);

    #
    # Connect the consumers
    #
    my @proxyIDs;
    my @processList;
    my @perProcessDataFiles;
    for(my $i = 0 ; $i < $perfOpts{'numConsumers'} ; $i++)
    {
	my $perProcessDataFile = "$outDataFile.c$i";
	push(@perProcessDataFiles, $perProcessDataFile);

	my $otherArgs = " --output-file $perProcessDataFile" .
			" -ORBconfig ob.conf";
	
	my $process = &forkProcess($consumerExe, "$consumerArgs $otherArgs");
	push(@processList, $process);
	sleep(1);
    }

    #
    # Start the suppliers
    #
    for(my $i = 0 ; $i < $perfOpts{'numSuppliers'} ; $i++)
    {
	my $perProcessDataFile = "$outDataFile.s$i";
	push(@perProcessDataFiles, $perProcessDataFile);

	my $otherArgs = " --output-file $perProcessDataFile" .
			" -ORBconfig ob.conf";

	my $process = &forkProcess($supplierExe, "$supplierArgs $otherArgs");
	push(@processList, $process);
    }

    #
    # Wait for the suppliers and consumers to go away
    #
    foreach my $process (@processList)
    {
	waitForProcess($process);
    }

    #
    # Stop the Event service
    #
    &shutdownService();

    #
    # Gather our results into a single file to make post processing a little
    # easier.
    #
    unlink("$outDataFile");
    open(DATA_FILE, ">$outDataFile");

    print DATA_FILE "Total number of events sent: $totalEvents\n";

    foreach my $perProcessDataFile (@perProcessDataFiles)
    {
	if(open(PER_PROCESS_DATA_FILE, $perProcessDataFile))
	{
	    while(<PER_PROCESS_DATA_FILE>)
	    {
		print DATA_FILE $_;
	    }
	    close(PER_PROCESS_DATA_FILE);
	}
    }
    close(DATA_FILE);

    sleep(1);
}

#
# Process & display test results.
#
sub postProcess(@)
{
    my @dataFiles = @_;

    foreach my $dataFile (@dataFiles)
    {
	my $outDataFile = getOutputFileName($dataFile);
	my @testResultData = &loadData("$outDataFile");
	my @supplierStats;
	my @consumerStats;
	my $totalEvents;

	foreach my $testResult (@testResultData)
	{
	    #
	    # Parse data
	    #
	    if($testResult =~
		/^First event sent by supplier: (\d?[-\+e\.\d]*)$/)
	    {
		push(@supplierStats, $1);
	    }
	    elsif($testResult =~
		/^Last event received by consumer: (\d?[-+e\.\d]*)$/)
	    {
		push(@consumerStats, $1);
	    }
	    elsif($testResult =~
		/^Total number of events sent: (\d?[-\+e\.\d]*)$/)
	    {
	        $totalEvents = $1;
	    }
	    else
	    {
		warn("Invalid data: \"$testResult\"");
	    }
	}

	#&printDataFile("$outDataFile");
	#&logIt("\n");

	#
	# Calculate and display results
	#
	my $firstEventSent = &getMinimum(@supplierStats);
	my $lastEventReceived = &getMaximum(@consumerStats);
	my $elapsedTime = $lastEventReceived - $firstEventSent;
	my $eventsPerSecond = ($elapsedTime > 0)
	    ? $totalEvents / $elapsedTime
	    : "infinite";
	#&logIt("    Last event received:   $lastEventReceived\n"); 
	#&logIt("    First event sent:      $firstEventSent\n"); 
	&logIt("    Number of events sent: $totalEvents\n");
	&logItf("    Time to send events:   %.4f seconds\n", $elapsedTime);
	&logItf("    Throughput:            %.4f events/second\n",
		$eventsPerSecond);
	&logIt("\n");
    }
}

sub printBanner($)
{
    my $dataFile = shift;

    my @inputData = &loadData($dataFile);

    &logIt("$dataFile\n");

    for(@inputData)
    {
	&logIt("    $_\n");
    }

    &logIt("\n");
}

sub printDataFile($)
{
    my @data = loadData(shift);

    for(@data)
    {
	&logIt("    $_\n");
    }
}

sub loadData($)
{
    my $dataFile = shift;

    if(!open(DATAFILE, $dataFile))
    {
	print "Unable to open $dataFile\n"; return;
	return;
    }

    my @data;
    while(<DATAFILE>)
    {
	chomp;
	push(@data, $_);
    }

    close(DATAFILE);

    return @data;
}

sub getMaximum(@)
{
    my @data = @_;
    my $max = 0;

    if($#data < 0)
    {
        return 0;
    }

    for(@data)
    {
        if($max == 0)
	{
	    $max = $_;
	}
	elsif($_ > $max)
	{
	    $max = $_;
	}
    }

    return $max;
}

sub getMinimum(@)
{
    my @data = @_;
    my $min = 0;

    if($#data < 0)
    {
        return 0;
    }

    for(@data)
    {
        if($min == 0)
	{
	    $min = $_;
	}
	elsif($_ < $min)
	{
	    $min = $_;
	}
    }

    return $min;
}

sub logIt($)
{
    my $logEntry = shift;

    print $logEntry;
    print LOG $logEntry;
}

sub logItf(@)
{
    my ($format, @arguments) = @_;

    printf $format, @arguments;
    printf LOG $format, @arguments;
}

sub getOutputFileName($)
{
    my $outDataFile = shift;

    $outDataFile =~ s/(.*)\.in/$1/;
    $outDataFile .= $OUT;

    return $outDataFile;
}

sub forkProcess($)
{
    my $command = shift;
    my $args = shift;
    my $processFd = new FileHandle;
    open($processFd, "$command $args|");

    return $processFd;
}

sub waitForProcess($)
{
    my $processFd = shift;
    while(<$processFd>)
    {
        sleep(1);
    }
}
