// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OB;

final class GIOPClientWorkerThreaded extends GIOPClientWorker
{
    protected int connectionTimeout_; // The connection timeout
    protected long timestamp_; // Timestamp for the last connection activity
    // Mutex for send
    protected java.lang.Object sendMutex_ = new java.lang.Object();

    protected final static class ShutdownThread extends Thread
    {
        private GIOPClientWorkerThreaded worker_;
        
        ShutdownThread(ThreadGroup group, GIOPClientWorkerThreaded worker)
        {
            super(group, "ORBacus:Client:ShutdownThread");
            worker_ = worker;
        }
        
        public void
        run()
        {
            try
            {
                worker_.shutdownRun();
            }
            catch(RuntimeException ex)
            {
                ex.printStackTrace();
                Assert._OB_assert(false);
            }
            //
            // Break cyclic object dependency
            //
            worker_ = null;
        }
    }

    protected Thread shutdownThread_;
    
    //
    // The receiver thread
    //
    protected final static class ReceiverThread extends Thread
    {
	private GIOPClientWorkerThreaded worker_;

	ReceiverThread(ThreadGroup group, GIOPClientWorkerThreaded worker)
        {
	    super(group, "ORBacus:Client:ReceiverThread");
	    worker_ = worker;
        }
        
	public void
        run()
	{
            try
            {
                worker_.receiverRun();
            }
            catch(RuntimeException ex)
            {
                ex.printStackTrace();
                Assert._OB_assert(false);
            }

            //
            // Break cyclic object dependency
            //
            worker_ = null;
	}
    }
    protected Thread receiverThread_;

    // ----------------------------------------------------------------------
    // GIOPClientWorkerThreaded private and protected member implementations
    // ----------------------------------------------------------------------

    //
    // Add new unsent buffer
    //
    protected void
    add(com.ooc.OCI.Buffer buf)
    {
        updateTimestamp();
        super.add(buf);
    }

    //
    // Add new unsent downcall
    //
    protected void
    add(Downcall down)
    {
        if(down.responseExpected())
            down.initStateMonitor();
        updateTimestamp();
        super.add(down);
    }

    protected void
    gracefulShutdown()
    {
        try
        {
            //
            // Retrieve the thread group for the clients
            //
            ThreadGroup group = orbInstance_.getClientWorkerGroup();
            
            //
            // Start the shutdown thread, which will take care of proper
            // connection termination
            //
            shutdownThread_ = new ShutdownThread(group, this);
            shutdownThread_.start();
        }
        catch(OutOfMemoryError ex)
        {
            exception(StateClosed, 
                      new org.omg.CORBA.IMP_LIMIT(
                          MinorCodes.describeImpLimit(
                              MinorCodes.MinorThreadLimit),
                          MinorCodes.MinorThreadLimit,
                          org.omg.CORBA.CompletionStatus.COMPLETED_NO));
        }
        return;
    }

    protected void
    abortiveShutdown()
    {    
        try
        {
            //
            // Send a MessageError message
            //
            com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer(12);
            com.ooc.CORBA.OutputStream out = 
                new com.ooc.CORBA.OutputStream(buf);
            com.ooc.OCI.ProfileInfo profileInfo = 
                new com.ooc.OCI.ProfileInfo();
            profileInfo.major = 1;
            profileInfo.minor = minor_;
            
            GIOPOutgoingMessage outgoing =
                new GIOPOutgoingMessage(orbInstance_, out, profileInfo);
            
            
            outgoing.writeMessageHeader(
                org.omg.GIOP.MsgType_1_1.MessageError, false, 0);
            out._OB_pos(0);
            transport_.send(out._OB_buffer(), true);
            Assert._OB_assert(out._OB_buffer().is_full());
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            exception(StateClosed, ex);
            return;
        }

        //
        // If we are in StateError, we don't go through all the hula hoop
        // with continuing to receive messages until the peer
        // closes. Instead, we just close the connection, meaning that we
        // can't be 100% sure that the peer gets the last message.
        //
        exception(StateClosed,
                  new org.omg.CORBA.TRANSIENT(
                      MinorCodes.describeTransient(
                          MinorCodes.MinorForcedShutdown),
                      MinorCodes.MinorForcedShutdown,
                      org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
    }

    //
    // Update the timestamp for the last connection activity
    //
    protected void
    updateTimestamp()
    {
        //
        // Only update if a connection timeout is set
        //
        if(connectionTimeout_ > 0)
        {
            timestamp_ = System.currentTimeMillis() / 1000;
        }
    }


    // ----------------------------------------------------------------------
    // GIOPClientWorkerThreaded package member implementations
    // ----------------------------------------------------------------------

    GIOPClientWorkerThreaded(ORBInstance orbInstance,
                             GIOPClient client,
			     com.ooc.OCI.Transport transport)
    {
	super(orbInstance, client, transport);

        //
        // Cache properties
        //
        java.util.Properties properties = orbInstance_.getProperties();
        String value;
        value = properties.getProperty("ooc.orb.client_timeout");
        if(value != null)
        {
            connectionTimeout_ = Integer.parseInt(value);
            updateTimestamp();
        }
        else
        {
            connectionTimeout_ = 0;
            timestamp_ = 0;
        }

        //
        // Only start a receiver thread if the transport supports
        // twoway invocations
        //
        if(transport.mode() == com.ooc.OCI.SendReceiveMode.SendReceive)
        {
            try
            {
                //
                // Retrieve the thread group for the clients
                //
                ThreadGroup group = orbInstance_.getClientWorkerGroup();

                //
                // Start receiver thread
                //
                receiverThread_ = new ReceiverThread(group, this);
                receiverThread_.start();
            }
            catch(OutOfMemoryError ex)
            {
                transport_.close();
                state_ = StateClosed;
                throw new org.omg.CORBA.IMP_LIMIT(
                    MinorCodes.describeImpLimit(MinorCodes.MinorThreadLimit),
                    MinorCodes.MinorThreadLimit,
                    org.omg.CORBA.CompletionStatus.COMPLETED_NO);
            }
        }
    }

    // ----------------------------------------------------------------------
    // GIOPClientWorkerThreaded public member implementations
    // ----------------------------------------------------------------------

    //
    // Destroy the worker
    //
    synchronized void 
    destroy()
    {
        exception(StateClosing,
                  new org.omg.CORBA.INITIALIZE(
                      MinorCodes.describeInitialize(
                          MinorCodes.MinorORBDestroyed),
                      MinorCodes.MinorORBDestroyed,
                      org.omg.CORBA.CompletionStatus.COMPLETED_NO));
    }

    //
    // Called to emit downcalls
    //
    public boolean
    send(Downcall down, boolean block)
    {
	if(block)
	{
	    //
	    // The following code is carrefully written in order to not
	    // lock the worker state when calling the send() method in
	    // blocking mode. We first get a message to send locking the
	    // worker state (we assume that the message will successfully
	    // be sent) and then we send the message without locking the
	    // worker state.
	    //
	    
	    boolean downcallAdded = false;
	    Downcall downcallToSend = null;

	    //
	    // Get the timeout
	    //
	    int t = down.policies().requestTimeout;
            
	    //
	    // Try to send the request
	    //
	    try
	    {
		do
		{
		    //
		    // Get a message to send from the unsent queue (and
		    // eventually add the downcall if not already done. We
		    // add the downcall here to avoid locking the worker
		    // state twice).
		    //
		    com.ooc.OCI.Buffer buf;
		    synchronized(this)
		    {
			if(!downcallAdded)
			{
			    add(down);
			    if(!down.unsent())
				return (!down.responseExpected());
			    downcallAdded = true;
			}

			buf = ((UnsentMessage)unsent_.firstElement()).buf;
			downcallToSend = moveFirstUnsentToPending();
		    }

		    //
		    // Send the message
		    //
		    synchronized(sendMutex_)
		    {
                        if(t < 0)
                        {
                            //
                            // Send buffer, blocking
                            //
                            transport_.send(buf, true);
                            Assert._OB_assert(buf.is_full());
                        }
                        else
                        {
                            //
                            // Send buffer, with timeout
                            //
                            transport_.send_timeout(buf, t);

                            // 
                            // Timeout?
                            // 
                            if(!buf.is_full())
                                throw new org.omg.CORBA.NO_RESPONSE(); 
                        }
		    }
		} 
		while(down.unsent());
	    }
	    catch(org.omg.CORBA.SystemException ex)
	    {
		synchronized(this)
		{
		    if(downcallToSend != null)
		    {
			//
			// Move  back the message to  the unsent queue
			// since we failed  to send it (this is mainly
			//  to have  an  accurate minor  code for  set
			// failure exception set on the downcall).
			//
			movePendingToUnsent(downcallToSend);
		    }
		    exception(StateClosed, ex);
		    return true;
		}
	    }
	}
	else // Non blocking
	{
	    synchronized(this)
	    {
		add(down);

		try
		{
		    while(down.unsent())
		    {
			com.ooc.OCI.Buffer buf
			    = ((UnsentMessage)unsent_.firstElement()).buf;
		
			synchronized(sendMutex_)
			{
			    //
			    // Send buffer, non-blocking
			    //
			    transport_.send(buf, false);
			    if(!buf.is_full())
				return false;
			}
			
			//
			// Message was sent completely
			//
			Downcall sent = moveFirstUnsentToPending();
                    }
                }
		catch(org.omg.CORBA.SystemException ex)
		{
		    exception(StateClosed, ex);
		    return true;
		}
	    }
	}
    
        if(down.responseExpected())
            return false;
        else
            return true;
    }
    
    public boolean
    receive(Downcall down, boolean block)
    {
        //
        // Try to receive the reply
        //
        try
        {
            return down.waitUntilCompleted(block);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            synchronized(this)
            {
                exception(StateClosed, ex);
                return true;
            }
        }
    }

    public boolean
    sendReceive(Downcall down)
    {
        if(send(down, true))
	    return true;

	return receive(down, true);
    }

    public void
    receiverRun()
    {
        while(true)
        {
            //
            // Setup the incoming message buffer
            //
            Assert._OB_assert(buf_ == null);
            buf_ = new com.ooc.OCI.Buffer(12);

            //
            // Receive header, blocking
            //
            try
            {
                transport_.receive(buf_, true);
                Assert._OB_assert(buf_.is_full());
            }
            catch(org.omg.CORBA.SystemException ex)
            {
		//
		// COMPILERFIX: The try/catch block is needed as a
		// workaround for the JDK 1.2.2 bug. Without this
		// workaround timeout tests fail with a
		// java.lang.IllegalMonitorStateException.
		//
		try
		{
		    synchronized(this)
		    {
			exception(StateClosed, ex);
			return;
		    }
		}
		catch(RuntimeException ex1)
		{
		    throw ex1;
	        }
            }

            //
            // Header is complete
            //
            try
            {
                incoming_.extractHeader(buf_);
                buf_.realloc(12 + incoming_.size());
            }
            catch(org.omg.CORBA.SystemException ex)
            {
		synchronized(this)
		{
		    exception(StateError, ex);
		    return;
		}
	    }

	    if(!buf_.is_full())
            {
                //
                // Receive body, blocking
                //
                try
                {
                    transport_.receive(buf_, true);
                    Assert._OB_assert(buf_.is_full());
                }
                catch(org.omg.CORBA.SystemException ex)
                {
		    synchronized(this)
		    {
			exception(StateClosed, ex);
			return;
		    }
                }
            }

            synchronized(this)
            {
                //
                // If we are not in StateActive or StateClosing, stop this
                // thread. We do *not* stop this thread if we are in
                // StateClosing, since we must continue to read data from
                // the Transport to make sure that no messages can get
                // lost upon close, and to make sure that CloseConnection
                // messages from the peer are processed.
                //
                if(state_ != StateActive && state_ != StateClosing)
                    return;
		
                //
                // Execute the message
                //
		execute();
            }
        }
    }

    public void
    shutdownRun()
    {
        if(minor_ >= 2)
        {
            //
            // Send a CloseConnection message
            //
            try
            {
                com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer(12);
                com.ooc.CORBA.OutputStream out = 
                    new com.ooc.CORBA.OutputStream(buf);
                com.ooc.OCI.ProfileInfo profileInfo = 
                    new com.ooc.OCI.ProfileInfo();
                profileInfo.major = 1;
                profileInfo.minor = minor_;
            
                GIOPOutgoingMessage outgoing =
                    new GIOPOutgoingMessage(orbInstance_, out, profileInfo);
                outgoing.writeMessageHeader(
                    org.omg.GIOP.MsgType_1_1.CloseConnection, false, 0);

                out._OB_pos(0);
                transport_.send(out._OB_buffer(), true);
                Assert._OB_assert(out._OB_buffer().is_full());
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                synchronized(this)
                {
                    exception(StateClosed, ex);
                }
            }
	}

        //
        // Now we gracefully shutdown the connection. Synchronization
        // is necessary to avoid a race with the receiver thread.
        //
        synchronized(this)
        {
            if(state_ == StateClosing)
            {
                logCloseConnection();
                transport_.shutdown();
            }
        }

        if(transport_.mode() == com.ooc.OCI.SendReceiveMode.SendReceive)
        {
            while(true)
            {
                try
                {
                    if(shutdownTimeout_ > 0)
                    {
                        receiverThread_.join(shutdownTimeout_ * 1000);
                    }
                    else
                    {
                        receiverThread_.join();
                    }
                    break;
                }
                catch(InterruptedException ex)
                {
                    continue;
                } 
            }
        }

        //
        // We now close the connection actively, since it may still be 
        // open under certain circumstances. (For example, if the 
        // receiver thread has not terminated yet, or if the 
        // receiver thread sets the state to StateError before 
        // termination.)
        //
        {
            synchronized(this)
            {
                exception(
                    StateClosed,
                    new org.omg.CORBA.TRANSIENT(
                        MinorCodes.MinorForcedShutdown, 
                        org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
            }
        }
    }
}
