// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OB;

import com.ooc.PortableServer.*;

abstract class GIOPServerWorker implements UpcallReturn
{
    protected ORBInstance orbInstance_; // The ORBInstance
    protected com.ooc.OCI.Transport transport_; // The transport
    protected OAInterface oaInterface_; // The object adapter interface
    protected com.ooc.OCI.Buffer buf_; // The incoming message buffer
    protected GIOPIncomingMessage incoming_; // The incoming message
    protected CodeConverters codeConverters_; // The code converters
    protected java.util.Vector unsent_ =
        new java.util.Vector(); // The outgoing message buffers
    protected int upcallCount_; // Number of outstanding upcalls
    protected byte minor_; // The highest minor version seen in peer messages
    protected boolean closeLogged_; // True if connection closure was already
                                    // logged
    protected boolean closeConnection_; // True if there was a CloseConnection
                                        // message
    protected int shutdownTimeout_; // The shutdown timeout

    //
    // The worker's state
    //
    public static final int StateActive  = 0;
    public static final int StateHolding = 1;
    public static final int StateClosing = 2;
    public static final int StateError   = 3;
    public static final int StateClosed  = 4;

    protected int state_;

    private boolean destroyed_ = false;

    // ----------------------------------------------------------------------
    // GIOPServerWorker private and protected member implementations
    // ----------------------------------------------------------------------

    //
    // Emit a trace message when closing the connection
    //
    protected void
    logCloseConnection()
    {
        if(!closeLogged_)
        {
            CoreTraceLevels coreTraceLevels =
                orbInstance_.getCoreTraceLevels();
            if(coreTraceLevels.traceConnections() > 0)
            {
                com.ooc.OCI.TransportInfo info = transport_.get_info();
                String msg = "closing connection\n";
                msg += info.describe();
                if(minor_ >= 2 && !closeConnection_)
                {
                    msg += "\nclient disconnected without sending " +
                           "a `CloseConnection' message";
                }
                orbInstance_.getLogger().trace("incoming", msg);
            }

            closeLogged_ = true;
        }
    }

    //
    // Get the code converters
    //
    protected void
    getCodeConverters(org.omg.IOP.ServiceContext[] scl)
    {
        if(codeConverters_ == null)
        {
            for(int i = 0 ; i < scl.length ; i++)
            {
                if(scl[i].context_id == org.omg.IOP.CodeSets.value)
                {
                    org.omg.CONV_FRAME.CodeSetContextHolder codeSetContextH =
                        new org.omg.CONV_FRAME.CodeSetContextHolder();
                    CodeSetUtil.extractCodeSetContext(scl[i],
                                                      codeSetContextH);
                    org.omg.CONV_FRAME.CodeSetContext codeSetContext =
                        codeSetContextH.value;

		    CodeSetDatabase db = CodeSetDatabase.instance();

                    codeConverters_ = new CodeConverters();
                    codeConverters_.inputCharConverter = 
                        db.getConverter(orbInstance_.getNativeCs(),
					codeSetContext.char_data);
                    codeConverters_.outputCharConverter = 
                        db.getConverter(codeSetContext.char_data,
					orbInstance_.getNativeCs());
                    codeConverters_.inputWcharConverter = 
                        db.getConverter(orbInstance_.getNativeWcs(),
					codeSetContext.wchar_data);
                    codeConverters_.outputWcharConverter = 
                        db.getConverter(codeSetContext.wchar_data,
					orbInstance_.getNativeWcs());

                    CoreTraceLevels coreTraceLevels =
                        orbInstance_.getCoreTraceLevels();
                    if(coreTraceLevels.traceConnections() >= 2)
                    {
                        String msg = "receiving transmission code sets";
                        msg += "\nchar code set: ";
                        if(codeConverters_.inputCharConverter != null)
                            msg += codeConverters_.inputCharConverter.getFrom()
                                .description;
                        else
			{
			    if(codeSetContext.char_data == 0)
				msg += "none";
			    else
			    {
				CodeSetInfo info =
				    db.getCodeSetInfo(
					orbInstance_.getNativeCs());
				msg += info.description;
			    }
			}
                        msg += "\nwchar code set: ";
                        if(codeConverters_.inputWcharConverter != null)
                            msg += codeConverters_.inputWcharConverter.
				getFrom().description;
                        else
			{
			    if(codeSetContext.wchar_data == 0)
				msg += "none";
			    else
			    {
				CodeSetInfo info =
				    db.getCodeSetInfo(
					orbInstance_.getNativeWcs());
				msg += info.description;
			    }
			}
                        orbInstance_.getLogger().trace("incoming", msg);
                    }

                    break;
                }
            }
        }
    }

    //
    // Change the state of the worker due to an exception
    //
    protected void
    exception(int state, org.omg.CORBA.SystemException ex)
    {
        String msg = "communications error with client\n";
        msg += ex.getMessage();
        orbInstance_.getLogger().warning(msg);

	//
	// In case the GIOP version number was too high the MessageError
	// reply has to indicate the highest supported GIOP version.
	// This version is set temporarily, i.e. only for this particular
	// reply.
	//
	byte minor = minor_;
	if(ex.minor == MinorCodes.MinorVersion)
	    minor_ = 2;
	setStateNoSync(state);
	minor_ = minor;
    }

    // 
    // Shutdown the connection gracefully
    //
    protected abstract void
    gracefulShutdown();

    // 
    // Abortive connection shutdown
    //
    protected abstract void
    abortiveShutdown();

    //
    // Return an upcall
    //
    protected abstract void
        upcallReturn(com.ooc.OCI.Buffer buf);
    
    //
    // Change the state of the worker without synchronization
    //
    protected boolean
    setStateNoSync(int state)
    {
        //
        // Don't do anything if there is no state change
        //
        if(state_ == state)
            return false;

        //
        // It is not possible to transition backwards, except if we are
	// in holding state
        //
        if(state_ != StateHolding && state < state_)
            return false;

        //
        // Update the state
        //
        state_ = state;

        switch(state)
        {
        case StateActive:
	case StateHolding:

            //
            // There is nothing special to do
            //
            break;

        case StateClosing:

            //
            // Do a graceful shutdown
            //
            gracefulShutdown();
	    destroyed_ = true;

            break;

        case StateError:

            //
            // Do an abortive shutdown
            //
            abortiveShutdown();
	    destroyed_ = true;

            break;

        case StateClosed:

            //
            // Close the transport
            //
            logCloseConnection();
            transport_.close();
	    destroyed_ = true;

            break;
        }

        return true;
    }

    //
    // Execute incoming message
    //
    protected Upcall
    execute()
    {
        Assert._OB_assert(state_ == StateActive || state_ == StateClosing);

        boolean complete;
        try
        {
            //
            // Provide the buffer to the incoming message and reset buf_
            // to allow for reading the next message. A return value of
            // false indicates that a message is not available (i.e., a
            // fragment was received).
            //
            com.ooc.OCI.Buffer buf = buf_;
            buf_ = null;
            complete = incoming_.consumeBuffer(buf);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            exception(StateError, ex);
            return null;
        }

        //
        // Keep track of the highest GIOP minor version seen
        //
        byte minor = incoming_.version().minor;
        if(minor > minor_)
            minor_ = minor;

        //
        // If the message is not complete yet, there is nothing to do
        //
        if(!complete)
            return null;

        //
        // OK, message is complete - let's execute it
        //
        Upcall upcall = null;

        switch(incoming_.type().value())
        {
        case org.omg.GIOP.MsgType_1_1._Request:

            //
            // Only execute the request if we are in StateActive. If we
            // are in StateClosing, then the server signals the client to
            // reissue the request by sending a CloseConnection message.
            //
            if(state_ == StateActive)
                upcall = executeRequest();
            break;

        case org.omg.GIOP.MsgType_1_1._CancelRequest:

            //
            // Do nothing here (doing nothing is GIOP compliant)
            //
            break;

        case org.omg.GIOP.MsgType_1_1._LocateRequest:

            //
            // Only execute the locate request if we are in StateActive
            //
            if(state_ == StateActive)
                executeLocate();
            break;

        case org.omg.GIOP.MsgType_1_1._Reply:
        case org.omg.GIOP.MsgType_1_1._LocateReply:

            //
            // These messages may not be sent to a server
            //
            exception(StateError,
                      new org.omg.CORBA.COMM_FAILURE(
                          MinorCodes.describeCommFailure(
                              MinorCodes.MinorWrongMessage),
                          MinorCodes.MinorWrongMessage,
                          org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
            break;

        case org.omg.GIOP.MsgType_1_1._CloseConnection:

            //
            // This message may not be sent to a server in GIOP 1.0 or 1.1
            //
            if(incoming_.version().minor < 2)
                exception(StateClosed,
                          new org.omg.CORBA.COMM_FAILURE(
                              MinorCodes.describeCommFailure(
                                  MinorCodes.MinorWrongMessage) +
                                  ": invalid version for CloseConnection",
                              MinorCodes.MinorWrongMessage,
                              org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
            else
            {
                closeConnection_ = true;
                setStateNoSync(StateClosed);
            }
            break;

        case org.omg.GIOP.MsgType_1_1._MessageError:

            //
            // How should we react to MessageError messages? We just send
            // a MessageError back, and close the connection.
            //
            exception(StateError,
                      new org.omg.CORBA.COMM_FAILURE(
                          MinorCodes.describeCommFailure(
                              MinorCodes.MinorMessageError),
                          MinorCodes.MinorMessageError,
                          org.omg.CORBA.CompletionStatus.COMPLETED_MAYBE));
            break;

        case org.omg.GIOP.MsgType_1_1._Fragment:

            //
            // At this point there should be no fragments, only complete
            // messages
            //
            Assert._OB_assert(false);
            break;
        }

        return upcall;
    }

    //
    // Execute Request message
    //
    protected Upcall
    executeRequest()
    {
        Assert._OB_assert(state_ == StateActive);

        org.omg.CORBA.BooleanHolder response =
            new org.omg.CORBA.BooleanHolder();
        int reqId;
        org.omg.CORBA.StringHolder op = new org.omg.CORBA.StringHolder();
        org.omg.IOP.ServiceContextListHolder scl =
            new org.omg.IOP.ServiceContextListHolder();
        org.omg.GIOP.TargetAddressHolder target =
            new org.omg.GIOP.TargetAddressHolder();

        try
        {
            reqId = incoming_.readRequestHeader(response, target, op, scl);
            if(target.value.discriminator() != org.omg.GIOP.KeyAddr.value)
            {
                exception(StateError,
                          new org.omg.CORBA.NO_IMPLEMENT(
                              MinorCodes.describeNoImplement(
                                  MinorCodes.MinorNotSupportedByLocalObject),
                              MinorCodes.MinorNotSupportedByLocalObject,
                              org.omg.CORBA.CompletionStatus.COMPLETED_NO));
                return null;
            }
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            exception(StateError, ex);
            return null;
        }

        //
        // Setup upcall data
        //
        org.omg.GIOP.Version version = incoming_.version();
        com.ooc.OCI.ProfileInfo profileInfo = new com.ooc.OCI.ProfileInfo();
        profileInfo.major = version.major;
        profileInfo.minor = version.minor;
        profileInfo.key = target.value.object_key();

        com.ooc.CORBA.InputStream in = incoming_.input();
        getCodeConverters(scl.value);
        if(codeConverters_ != null)
            in._OB_codeConverters(codeConverters_,
                                  (version.major << 8) | version.minor);

        //
        // New upcall will be started
        //
        if(response.value)
            upcallCount_++;

        return oaInterface_.createUpcall(
            response.value ? this : null,
            profileInfo,
            transport_.get_info(),
            reqId,
            op.value,
            in,
            scl.value);
    }

    //
    // Execute LocateRequest message
    //
    protected void
    executeLocate()
    {
        Assert._OB_assert(state_ == StateActive);

        //
        // Make sure the transport can send a reply
        //
        if(transport_.mode() == com.ooc.OCI.SendReceiveMode.ReceiveOnly)
        {
            String msg = "Discarding locate request - transport " +
                         "does not support twoway invocations";

            com.ooc.OCI.TransportInfo transportInfo = transport_.get_info();
            if(transportInfo != null)
            {
                String desc = transportInfo.describe();
                msg += '\n';
                msg += desc;
            }
            else
            {
                msg += "\nCollocated method call";
            }

            Logger logger = orbInstance_.getLogger();
            logger.warning(msg);

            return;
        }

        try
        {
            int reqId;
            org.omg.GIOP.TargetAddressHolder target =
                new org.omg.GIOP.TargetAddressHolder();
            reqId = incoming_.readLocateRequestHeader(target);

            if(target.value.discriminator() != org.omg.GIOP.KeyAddr.value)
            {
                exception(StateError,
                          new org.omg.CORBA.NO_IMPLEMENT(
                              MinorCodes.describeNoImplement(
                                  MinorCodes.MinorNotSupportedByLocalObject),
                              MinorCodes.MinorNotSupportedByLocalObject,
                              org.omg.CORBA.CompletionStatus.COMPLETED_NO));
                return;
            }

            //
            // Get the key
            //
            byte[] key = target.value.object_key();

            //
            // Find the IOR for the key
            //
            org.omg.IOP.IORHolder ior = new org.omg.IOP.IORHolder();
            int val = oaInterface_.findByKey(key, ior);
            org.omg.GIOP.LocateStatusType_1_2 status =
                org.omg.GIOP.LocateStatusType_1_2.from_int(val);

            //
            // Send back locate reply message
            //
            com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer(12);
            buf.pos(12);
            com.ooc.CORBA.OutputStream out =
                new com.ooc.CORBA.OutputStream(buf);
            com.ooc.OCI.ProfileInfo profileInfo =
                new com.ooc.OCI.ProfileInfo();
            profileInfo.major = incoming_.version().major;
            profileInfo.minor = incoming_.version().minor;
            GIOPOutgoingMessage outgoing =
                new GIOPOutgoingMessage(orbInstance_, out, profileInfo);
            outgoing.writeLocateReplyHeader(reqId, status);

            //
            // If the reply status is OBJECT_FORWARD or
            // OBJECT_FORWARD_PERM the IOR is appended to the end of the
            // LocateReply.
            //
            if(status == org.omg.GIOP.LocateStatusType_1_2.OBJECT_FORWARD ||
               status == org.omg.GIOP.LocateStatusType_1_2.OBJECT_FORWARD_PERM)
                org.omg.IOP.IORHelper.write(out, ior.value);

            //
            // TODO:
            // LOC_SYSTEM_EXCEPTION,
            // LOC_NEEDS_ADDRESSING_MODE
            //
            int pos = out._OB_pos();
            out._OB_pos(0);
            outgoing.writeMessageHeader(org.omg.GIOP.MsgType_1_1.LocateReply,
                                        false, pos - 12);
            out._OB_pos(pos);

            //
            // A locate request is treated just like an upcall
            //
            upcallCount_++;

            //
            // Send the locate reply
            //
            upcallReturn(out._OB_buffer());
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            exception(StateError, ex);
        }
    }

    protected void
    finalize()
        throws Throwable
    {
        Assert._OB_assert(state_ == StateClosed);

        super.finalize();
    }

    // ----------------------------------------------------------------------
    // GIOPServerWorker package member implementations
    // ----------------------------------------------------------------------

    GIOPServerWorker(ORBInstance orbInstance,
                     com.ooc.OCI.Transport transport,
                     OAInterface oaInterface)
    {
        orbInstance_ = orbInstance;
        transport_ = transport;
        oaInterface_ = oaInterface;
        incoming_ = new GIOPIncomingMessage(orbInstance_);
        codeConverters_ = null;
        upcallCount_ = 0;
        minor_ = 0;
        closeLogged_ = false;
        closeConnection_ = false;
        state_ = StateHolding;
	destroyed_ = false;

        //
        // Cache properties
        //
        java.util.Properties properties = orbInstance_.getProperties();
        String value;

        value = properties.getProperty("ooc.orb.server_shutdown_timeout");
        if(value != null)
            shutdownTimeout_ = Integer.parseInt(value);
        else
            shutdownTimeout_ = 2;

        //
        // Trace new incoming connection
        //
        CoreTraceLevels coreTraceLevels = orbInstance_.getCoreTraceLevels();
        if(coreTraceLevels.traceConnections() > 0)
        {
            com.ooc.OCI.TransportInfo info = transport_.get_info();
            String msg = "new connection\n";
            msg += info.describe();
            orbInstance_.getLogger().trace("incoming", msg);
        }
    }

    // ----------------------------------------------------------------------
    // GIOPServerWorker public member implementations
    // ----------------------------------------------------------------------

    //
    // Change the state of the worker
    //
    public abstract boolean
    setState(int state);

    public boolean
    destroyed()
    {
        return destroyed_;
    }

    public void
    upcallBeginReply(Upcall upcall, org.omg.IOP.ServiceContext[] replySCL)
    {
        upcall.createOutputStream(12);

        com.ooc.CORBA.OutputStream out = upcall.output();
        com.ooc.OCI.ProfileInfo profileInfo = upcall.profileInfo();
        GIOPOutgoingMessage outgoing =
            new GIOPOutgoingMessage(orbInstance_, out, profileInfo);

        int reqId = upcall.requestId();
        try
        {
            outgoing.writeReplyHeader(
                reqId, org.omg.GIOP.ReplyStatusType_1_2.NO_EXCEPTION,
                replySCL);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            //
            // Nothing may go wrong here, otherwise we might have a
            // recursion
            //
            Assert._OB_assert(false);
        }
    }

    public void
    upcallEndReply(Upcall upcall)
    {
        //
        // Make sure the transport can send a reply
        //
        if(transport_.mode() == com.ooc.OCI.SendReceiveMode.ReceiveOnly)
        {
            String msg = "Discarding reply - transport does not " +
                         "support twoway invocations";

            msg += "\noperation name: \"";
            msg += upcall.operation();
            msg += '"';

            com.ooc.OCI.TransportInfo transportInfo = transport_.get_info();
            if(transportInfo != null)
            {
                String desc = transportInfo.describe();
                msg += '\n';
                msg += desc;
            }
            else
            {
                msg += "\nCollocated method call";
            }

            Logger logger = orbInstance_.getLogger();
            logger.warning(msg);

            return;
        }

        com.ooc.CORBA.OutputStream out = upcall.output();
        com.ooc.OCI.ProfileInfo profileInfo = upcall.profileInfo();
        GIOPOutgoingMessage outgoing =
            new GIOPOutgoingMessage(orbInstance_, out, profileInfo);

        int pos = out._OB_pos();
        out._OB_pos(0);
        try
        {
            outgoing.writeMessageHeader(org.omg.GIOP.MsgType_1_1.Reply, false,
                                        pos - 12);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            //
            // Nothing may go wrong here, otherwise we might have a
            // recursion
            //
            Assert._OB_assert(false);
        } 

        upcallReturn(out._OB_buffer());
    }

    public void
    upcallBeginUserException(Upcall upcall,
                             org.omg.IOP.ServiceContext[] replySCL)
    {
        upcall.createOutputStream(12);

        com.ooc.CORBA.OutputStream out = upcall.output();
        com.ooc.OCI.ProfileInfo profileInfo = upcall.profileInfo();
        GIOPOutgoingMessage outgoing =
            new GIOPOutgoingMessage(orbInstance_, out, profileInfo);

        int reqId = upcall.requestId();

        try
        {
            outgoing.writeReplyHeader(
                reqId, org.omg.GIOP.ReplyStatusType_1_2.USER_EXCEPTION,
                replySCL);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            //
            // Nothing may go wrong here, otherwise we might have a
            // recursion
            //
            Assert._OB_assert(false);
        }
    }

    public void
    upcallEndUserException(Upcall upcall)
    {
        upcallEndReply(upcall);
    }

    //
    // NOTE: Not used in Java
    //
    public void
    upcallUserException(Upcall upcall,
                        org.omg.CORBA.UserException ex,
                        org.omg.IOP.ServiceContext[] replySCL)
    {
        upcall.createOutputStream(12);

        com.ooc.CORBA.OutputStream out = upcall.output();
        com.ooc.OCI.ProfileInfo profileInfo = upcall.profileInfo();
        GIOPOutgoingMessage outgoing =
            new GIOPOutgoingMessage(orbInstance_, out, profileInfo);

        int reqId = upcall.requestId();

        try
        {
            outgoing.writeReplyHeader(
                reqId, org.omg.GIOP.ReplyStatusType_1_2.USER_EXCEPTION,
                replySCL);
            //
            // Cannot marshal the exception without the Helper
            //
            //ex._OB_marshal(out);
            Assert._OB_assert(false);
        }
        catch(org.omg.CORBA.SystemException e)
        {
            //
            // Nothing may go wrong here, otherwise we might have a
            // recursion
            //
            Assert._OB_assert(false);
        }

        upcallEndReply(upcall);
    }

    public void
    upcallSystemException(Upcall upcall, org.omg.CORBA.SystemException ex,
                          org.omg.IOP.ServiceContext[] replySCL)
    {
        upcall.createOutputStream(12);

        com.ooc.CORBA.OutputStream out = upcall.output();
        com.ooc.OCI.ProfileInfo profileInfo = upcall.profileInfo();
        GIOPOutgoingMessage outgoing =
            new GIOPOutgoingMessage(orbInstance_, out, profileInfo);

        int reqId = upcall.requestId();
        try
        {
            outgoing.writeReplyHeader(
                reqId, org.omg.GIOP.ReplyStatusType_1_2.SYSTEM_EXCEPTION,
                replySCL);
            Util.marshalSystemException(out, ex);
        }
        catch(org.omg.CORBA.SystemException e)
        {
            //
            // Nothing may go wrong here, otherwise we might have a
            // recursion
            //
            Assert._OB_assert(false);
        }

        upcallEndReply(upcall);
    }

    public void
    upcallForward(Upcall upcall, org.omg.IOP.IOR ior, boolean perm,
                  org.omg.IOP.ServiceContext[] replySCL)
    {
        upcall.createOutputStream(12);

        com.ooc.CORBA.OutputStream out = upcall.output();
        com.ooc.OCI.ProfileInfo profileInfo = upcall.profileInfo();
        GIOPOutgoingMessage outgoing =
            new GIOPOutgoingMessage(orbInstance_, out, profileInfo);

        int reqId = upcall.requestId();
        org.omg.GIOP.ReplyStatusType_1_2 status =
            perm ? org.omg.GIOP.ReplyStatusType_1_2.LOCATION_FORWARD_PERM
                 : org.omg.GIOP.ReplyStatusType_1_2.LOCATION_FORWARD;
        try
        {
            outgoing.writeReplyHeader(reqId, status, replySCL);
            org.omg.IOP.IORHelper.write(out, ior);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            //
            // Nothing may go wrong here, otherwise we might have a
            // recursion
            //
            Assert._OB_assert(false);
        }

        upcallEndReply(upcall);
    }
}
