// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

package com.ooc.OB;

final class GIOPServerWorkerThreaded extends GIOPServerWorker
{
    //
    // The sender thread
    //
    protected int connectionTimeout_; // The connection timeout
    protected long timestamp_; // Timestamp for the last connection activity

    protected final static class ShutdownThread extends Thread
    {
        private GIOPServerWorkerThreaded worker_;
        
        ShutdownThread(ThreadGroup group, GIOPServerWorkerThreaded worker)
        {
            super(group, "ORBacus:Server:ShutdownThread");
            worker_ = worker;
        }
        
        public void
        run()
        {
            try
            {
                worker_.shutdownRun();
            }
            catch(RuntimeException ex)
            {
                ex.printStackTrace();
                Assert._OB_assert(false);
            }
            //
            // Break cyclic object dependency
            //
            worker_ = null;
        }
    }

    protected Thread shutdownThread_;

    //
    // The receiver thread
    //
    protected final static class ReceiverThread extends Thread
    {
        private GIOPServerWorkerThreaded worker_;

        ReceiverThread(ThreadGroup group, GIOPServerWorkerThreaded worker)
        {
            super(group, "ORBacus:Server:ReceiverThread");
            worker_ = worker;
        }

        public void
        run()
        {
            try
            {
                worker_.receiverRun();
            }
            catch(RuntimeException ex)
            {
                ex.printStackTrace();
                Assert._OB_assert(false);
            }

            //
            // Break cyclic object dependency
            //
            worker_ = null;
        }
    }
    protected Thread receiverThread_;

    // ----------------------------------------------------------------------
    // GIOPServerWorkerThreaded private and protected member implementations
    // ----------------------------------------------------------------------

    //
    // Change the state of the worker without synchronization
    //
    protected boolean
    setStateNoSync(int state)
    {
        //
        // Let the base class change the state
        //
        if(!super.setStateNoSync(state))
            return false;

        //
        // Notify about the state change
        //
        notify();
        return true;
    }

    protected void
    gracefulShutdown()
    { 
        //
        // Don't shut down if there are still outstanding upcalls
        //
        if(upcallCount_ > 0)
            return;

        try
        {
            //
            // Retrieve the thread group for the servers
            //
            ThreadGroup group = orbInstance_.getServerWorkerGroup();
            
            //
            // Start the shutdown thread, which will take care of proper
            // connection termination
            //
            shutdownThread_ = new ShutdownThread(group, this);
            shutdownThread_.start();
        }
        catch(OutOfMemoryError ex)
        {
            exception(StateClosed, 
                      new org.omg.CORBA.IMP_LIMIT(
                          MinorCodes.describeImpLimit(
                              MinorCodes.MinorThreadLimit),
                          MinorCodes.MinorThreadLimit,
                          org.omg.CORBA.CompletionStatus.COMPLETED_NO));
        }
    }
    
    protected void
    abortiveShutdown()
    {    
        if(transport_.mode() == com.ooc.OCI.SendReceiveMode.SendReceive)
        {
            try
            {
                //
                // Send a MessageError message
                //
                com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer(12);
                com.ooc.CORBA.OutputStream out = 
                    new com.ooc.CORBA.OutputStream(buf);
                com.ooc.OCI.ProfileInfo profileInfo = 
                    new com.ooc.OCI.ProfileInfo();
                profileInfo.major = 1;
                profileInfo.minor = minor_;
                
                GIOPOutgoingMessage outgoing =
                    new GIOPOutgoingMessage(orbInstance_, out, profileInfo);
                
                outgoing.writeMessageHeader(
                    org.omg.GIOP.MsgType_1_1.MessageError, false, 0);
                out._OB_pos(0);
                if(!transport_.send_detect(out._OB_buffer(), true))
                {
                    setStateNoSync(StateClosed);
                    return;
                }
                Assert._OB_assert(out._OB_buffer().is_full());
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                exception(StateClosed, ex);
                return;
            }
        }

        //
        // If we are in StateError, we don't go through all the hula hoop
        // with continuing to receive messages until the peer
        // closes. Instead, we just close the connection, meaning that we
        // can't be 100% sure that the peer gets the last message.
        //
        setStateNoSync(StateClosed);
    }

    //
    // Update the timestamp for the last connection activity
    //
    protected void
    updateTimestamp()
    {
        //
        // Only update if a connection timeout is set
        //
        if(connectionTimeout_ > 0)
        {
            timestamp_ = System.currentTimeMillis() / 1000;
        }
    }

    protected void
    upcallReturn(com.ooc.OCI.Buffer buf)
    {
	//
	// We need synchronization here since this method might be
	// called concurrently by different threads (the dispatching
	// threads for instance, i.e.: the thread pool or the request
	// thread).
	//
        synchronized(this)
        {
	    buf.pos(0);
	    try
	    {
		if(!transport_.send_detect(buf,true))
		{
		    setState(StateClosed);
		    return;
		}
		Assert._OB_assert(buf.is_full());
	    } 
	    catch(org.omg.CORBA.SystemException ex)
	    {
		exception(StateClosed, ex);
		return;
	    }

            Assert._OB_assert(upcallCount_ > 0);
            upcallCount_--;

            //
            // If this was the last upcall and we are in StateClosing, 
            // we now gracefully shut down the connection
            //
            if(upcallCount_ == 0 && state_ == StateClosing)
                gracefulShutdown();
        }
    }      


    // ----------------------------------------------------------------------
    // GIOPServerWorkerThreaded package member implementations
    // ----------------------------------------------------------------------

    GIOPServerWorkerThreaded(
        ORBInstance orbInstance,
        com.ooc.OCI.Transport transport,
        OAInterface oaInterface)
    {
        super(orbInstance, transport, oaInterface);

        //
        // Cache properties
        //
        java.util.Properties properties = orbInstance_.getProperties();
        String value;

        value = properties.getProperty("ooc.orb.server_timeout");
        if(value != null)
        {
            connectionTimeout_ = Integer.parseInt(value);
            updateTimestamp();
        }
        else
        {
            connectionTimeout_ = 0;
            timestamp_ = 0;
        }

        try
        {
            //
            // Retrieve the thread group for the servers
            //
            ThreadGroup group = orbInstance_.getServerWorkerGroup();

            //
            // Start receiver thread
            //
            receiverThread_ = new ReceiverThread(group, this);
            receiverThread_.start();
        }
        catch(OutOfMemoryError ex)
        {
            transport_.close();
            state_ = StateClosed;
            throw new org.omg.CORBA.IMP_LIMIT(
                MinorCodes.describeImpLimit(MinorCodes.MinorThreadLimit),
                MinorCodes.MinorThreadLimit,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    // ----------------------------------------------------------------------
    // GIOPServerWorkerThreaded public member implementations
    // ----------------------------------------------------------------------

    public synchronized boolean
    setState(int state)
    {
        return setStateNoSync(state);
    }

    public void
    receiverRun()
    {
        while(true)
        {
            //
            // Setup the incoming message buffer
            //
            Assert._OB_assert(buf_ == null);
            buf_ = new com.ooc.OCI.Buffer(12);

            //
            // Receive header, blocking, detect connection loss
            //
            try
            {
                while(!buf_.is_full())
                {
                    if(connectionTimeout_ > 0)
                    {
                        int oldPos = buf_.pos();
                        
                        // 
                        // Receive header, with timeout, detect 
                        // connection loss
                        // 
                        if(!transport_.
                           receive_timeout_detect(buf_, 
                                                  connectionTimeout_ * 1000))
                        {
                            synchronized(this)
                            {
                                setStateNoSync(StateClosed);
                            }
                            return; // Orderly shutdown
                        }
                        
                        //
                        // No data received at all?
                        //
                        if(buf_.pos() == oldPos)
                        {
                            synchronized(this)
                            {
                                //
                                // Shut down if there are no outstanding
                                // upcalls and the time limit has been
                                // exceeded
                                //
                                if(upcallCount_ == 0)
                                {		    
                                    long current = 
                                        System.currentTimeMillis()/1000;
                                    
                                    if(current >= 
                                       timestamp_ + connectionTimeout_)
                                    {
                                        connectionTimeout_ = 0;
                                        setStateNoSync(StateClosing);
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        //
                        // Receive header, blocking, detect connection loss
                        // 
                        if(!transport_.receive_detect(buf_, true))
                        {
                            synchronized(this)
                            {
                                setStateNoSync(StateClosed);
                            }
                            return; // Orderly shutdown
                        }
                        Assert._OB_assert(buf_.is_full());
                    }
                }
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                synchronized(this)
                {
                    exception(StateClosed, ex);
                }
                return;
            }
                
            //
            // Header is complete
            //
            try
            {
                incoming_.extractHeader(buf_);
                buf_.realloc(12 + incoming_.size());
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                synchronized(this)
                {
                    exception(StateError, ex);
                }
                return;
            }
                
            try
            {
                while(!buf_.is_full())
                {
                    if(connectionTimeout_ > 0)
                    {
                        long oldPos = buf_.pos();
                            
                        //
                        // Receive body, with timeout, detect connection
                        // loss
                        //
                        if(!transport_.
                           receive_timeout_detect(buf_, 
                                                  connectionTimeout_ * 1000))
                        {
                            synchronized(this)
                            {
                                setStateNoSync(StateClosed);
                            }
                            return; // Orderly shutdown
                        }
                            
                        //
                        // No data received at all?
                        //
                        if(buf_.pos() == oldPos)
                        {
                            synchronized(this)
                            {
                                //
                                // Shut down if there are no outstanding
                                // upcalls and the time limit has been
                                // exceeded
                                //
                                if(upcallCount_ == 0)
                                {		    
                                    long current = 
                                        System.currentTimeMillis()/1000;
                                    
                                    if(current >= 
                                       timestamp_ + connectionTimeout_)
                                    {
                                        connectionTimeout_ = 0;
                                        setStateNoSync(StateClosing);
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        //
                        // Receive body, blocking, detect connection loss
                        //
                        if(!transport_.receive_detect(buf_, true))
                        {
                            synchronized(this)
                            {
                                setStateNoSync(StateClosed);
                            }
                            return; // Orderly shutdown
                        }
                        Assert._OB_assert(buf_.is_full());
                    }
                }
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                synchronized(this)
                {
                    exception(StateClosed, ex);
                }
                return;
            }
            
            Upcall upcall;
            
            synchronized(this)
            {
	        //
		// Wait until we are not holding anymore
		//
                //
		while(state_ == StateHolding)
		{
		    try
		    {
		        wait();
		    }
		    catch(InterruptedException ex)
		    {
		    }
		}

                // If we are not in StateActive or StateClosing, stop this
                // thread. We do *not* stop this thread if we are in
                // StateClosing, since we must continue to read data from
                // the Transport to make sure that no messages can get
                // lost upon close, and to make sure that CloseConnection
                // messages from the peer are processed.
                //
                if(state_ != StateActive && state_ != StateClosing)
                    return;
                
                //
                // Execute the message
                //
                updateTimestamp();
                upcall = execute();
            }
                
            //
            // Invoke the upcall, if there is one
            //
            if(upcall != null)
                upcall.invoke();
        }
    }

    public void
    shutdownRun()
    {
        closeConnection_ = true;

        if(transport_.mode() == com.ooc.OCI.SendReceiveMode.SendReceive)
        {
            //
            // Send a CloseConnection message
            //
            try
            {
                com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer(12);
                com.ooc.CORBA.OutputStream out = 
                    new com.ooc.CORBA.OutputStream(buf);
                com.ooc.OCI.ProfileInfo profileInfo = 
                    new com.ooc.OCI.ProfileInfo();
                profileInfo.major = 1;
                profileInfo.minor = minor_;
                
                GIOPOutgoingMessage outgoing =
                    new GIOPOutgoingMessage(orbInstance_, out, profileInfo);
                outgoing.writeMessageHeader(
                    org.omg.GIOP.MsgType_1_1.CloseConnection, false, 0);
                
                out._OB_pos(0);
                if(!transport_.send_detect(out._OB_buffer(), true))
                {
                    synchronized(this)
                    {
                        setStateNoSync(StateClosed);
                        return; // Orderly shutdown
                    }
                }
                Assert._OB_assert(out._OB_buffer().is_full());
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                synchronized(this)
                {
                    exception(StateClosed, ex);
                    return;
                }
            }
        }

        //
        // Now we gracefully shutdown the connection. Synchronization
        // is necessary to avoid a race with the receiver thread.
        //
        synchronized(this)
        {
            if(state_ == StateClosing)
            {
                logCloseConnection();
                transport_.shutdown();
            }
        }

        if(transport_.mode() == com.ooc.OCI.SendReceiveMode.SendReceive)
        {
            while(true)
            {
                try
                {
                    if(shutdownTimeout_ > 0)
                    {
                        //
                        // Join with the receiver thread, with timeout
                        //
                        receiverThread_.join(shutdownTimeout_ * 1000);
                    }
                    else
                    {
                        //
                        // Join with the receiver thread, no timeout
                        //
                        receiverThread_.join();
                    }

                    break;
                }
                catch(InterruptedException ex)
                {
                    continue;
                } 
            }
        }

        //
        // We now close the connection actively, since it may still be
        // open under certain circumstances. (For example, if the receiver
        // thread has not terminated yet, or if the receiver thread sets
        // the state to StateError before termination.)
        //
        synchronized(this)
        {
            setStateNoSync(StateClosed);
        }
    }
}
